/**
 ** @file mod_idmef.c
 ** Listen to idmef alerts on a udp socket.
 **
 ** @author Baptiste GOURDIN <gourdin@lsv.ens-cachan.fr>
 **
 ** @version 0.1
 ** @ingroup modules
 **
 **/

/*
 * See end of file for LICENSE and COPYRIGHT informations.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <ctype.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <libxml/xmlschemas.h>


#include "mod_idmef.h"
#include "mod_xml.h"

input_module_t mod_idmef;

field_t idmef_fields[MAX_IDMEF_FIELDS] = {
  { "idmef.ptr",		    T_PTR32,"idmef xml doc"   }
};

static ovm_var_t*
parse_idmef_datetime(char	*datetime)
{
  struct tm tm;
  char	*datetime_ = datetime;
  int	h;
  int	m;
  ovm_var_t	*res;


  memset(&tm, 0, sizeof(tm));
  tm.tm_isdst = -1;

  datetime = strptime(datetime, "%Y-%m-%dT%H:%M:%S", &tm);

  if (!datetime)
  {
    DebugLog(DF_MOD, DS_ERROR, "Malformated idmef datetime (%s)\n", datetime_);
    return NULL;
  }

  if ((*datetime == '.') ||
      (*datetime == ','))
  {
    // skip fraction seconds
    datetime++;
    while ( isdigit(*datetime) )
      datetime++;
  }

  if ((*datetime == '+') ||
      (*datetime == '-'))
  {
    if (sscanf(datetime + 1, "%2u:%2u", &h, &m) != 2)
      DebugLog(DF_MOD, DS_ERROR, "Malformated idmef datetime (%s)\n", datetime_);
    if (*datetime == '+')
    {
      tm.tm_min -= m;
      tm.tm_hour -= h;
    }
    else
    {
      tm.tm_min += m;
      tm.tm_hour += h;
    }
  }

  res = ovm_ctime_new();
  CTIME(res) = mktime(&tm);
  return res;
}

static int
load_idmef_xmlDoc(orchids_t	*ctx,
		  mod_entry_t	*mod,
		  const char	*txt_line,
		  const size_t	txt_len)
{
  idmef_cfg_t	*cfg;
  xmlDocPtr	doc = NULL;
  ovm_var_t	*attr[MAX_IDMEF_FIELDS];
  event_t	*event;
  xmlXPathContextPtr	xpath_ctx = NULL;
  xml_doc_t	*xml_doc = NULL;
  int		c;

  cfg = mod->config;
  if (((doc = xmlReadMemory(txt_line, txt_len, "idmef", NULL, 0)) == NULL)
      || ((xpath_ctx = xmlXPathNewContext(doc)) == NULL))
  {
    fprintf(stdout, "Error loading IDMEF xml doc \n");
    return (0);
  }

  xml_doc = Xzmalloc(sizeof (xml_doc_t));
  xml_doc->doc = doc;
  xml_doc->xpath_ctx = xpath_ctx;

  xmlXPathRegisterNs(xpath_ctx, BAD_CAST ("idmef"),
		     BAD_CAST ("http://iana.org/idmef"));

  memset(attr, 0, sizeof(attr));
  attr[F_PTR] = ovm_ptr32_new();
  PTR32(attr[F_PTR]) = xml_doc;

  for (c = 1; c < cfg->nb_fields; c++)
  {
    char	*text;
    int		text_len;
    ovm_var_t	*res;

    if ((text = xml_get_string(xml_doc, cfg->field_xpath[c])) != NULL)
    {
      switch (idmef_fields[c].type)
      {
	case T_STR :
	case T_VSTR :
	  text_len = strlen(text);
	  res = ovm_str_new(text_len);
	  memcpy (STR(res), text, text_len);
	  attr[c] = res;
	  break;
	case T_CTIME :
	{
	  attr[c] = parse_idmef_datetime(text);
	  break;
	}
	case T_IPV4 :
	  attr[c] = ovm_ipv4_new();
	  IPV4(attr[c]).s_addr = inet_addr(text);
	  break;
	case T_INT :
	  attr[c] = ovm_int_new();
	  sscanf(text, "%li", &(INT(attr[c])));
	  break;
	case T_UINT :
	  attr[c] = ovm_uint_new();
	  sscanf(text, "%lu", &(UINT(attr[c])));
	  break;
      }
    }
  }

  event = NULL;
  add_fields_to_event(ctx, mod, &event, attr, cfg->nb_fields);
  post_event(ctx, mod, event);

  return (1);
}

static int
dissect_idmef(orchids_t *ctx, mod_entry_t *mod, event_t *event, void *data)
{
  char *line;
  char *end;
  static char	buff[8192] = {0};
  static int	buff_len;
  DebugLog(DF_MOD, DS_DEBUG, "idmef_dissector()\n");

  line = STR(event->value);

  while ((end = strstr(line, "</idmef:IDMEF-Message>")) != NULL)
  {
    if (buff_len +  end - line + 22 >= 8192)
    {
      DebugLog(DF_MOD, DS_ERROR, "idmef message too big\n");

    }
    else
    {
      strncat (buff, line, end - line + 22);
      load_idmef_xmlDoc(ctx, mod, buff, strlen(buff));
    }
    *buff = 0;
    line = end + 22;
  }

  strcat (buff, line);

  return (0);
}

static void *
idmef_preconfig(orchids_t *ctx, mod_entry_t *mod)
{
  idmef_cfg_t*	cfg;

  DebugLog(DF_MOD, DS_INFO, "load() idmef@%p\n", (void *) &mod_idmef);

  cfg = Xzmalloc(sizeof (idmef_cfg_t));
  cfg->nb_fields = 1;

  return (cfg);
}

static void
idmef_postconfig(orchids_t *ctx, mod_entry_t *mod)
{
  DebugLog(DF_MOD, DS_INFO, "load() idmef %i\n", ((idmef_cfg_t*)mod->config)->nb_fields );

  register_fields(ctx, mod, idmef_fields, ((idmef_cfg_t*)mod->config)->nb_fields);
}

static void
add_field(char* field_name, char* path, idmef_cfg_t*	cfg, int type)
{
  if (cfg->nb_fields == MAX_IDMEF_FIELDS)
  {
    DebugLog(DF_MOD, DS_WARN, "Max number of idmef fields reached, cannot add %s %s\n", field_name , path);
    return;
  }

  idmef_fields[cfg->nb_fields].name = Xzmalloc((9 + strlen(field_name))
							  * sizeof(char));
  strcat(idmef_fields[cfg->nb_fields].name, "idmef.");
  strcat(idmef_fields[cfg->nb_fields].name, field_name);

  idmef_fields[cfg->nb_fields].type = type;
  idmef_fields[cfg->nb_fields].desc = path;

  DebugLog(DF_MOD, DS_INFO, "add new field %i (%s) : %s\n",
	   cfg->nb_fields,
	   idmef_fields[cfg->nb_fields].name , path);
  cfg->field_xpath[cfg->nb_fields++] = path;
}

static void
dir_add_str_field(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  char* pos;

  for (pos = dir->args; *pos && !isblank(*pos); pos++)
    continue;
  if (!pos)
  {
    DebugLog(DF_MOD, DS_ERROR, "Error when parsing directive str_field\n");
    return;
  }
  *pos = '\0';

  add_field(dir->args, pos + 1, mod->config, T_STR);
}


static void
dir_add_time_field(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  char* pos;

  for (pos = dir->args; *pos && !isblank(*pos); pos++)
    continue;
  if (!pos)
  {
    DebugLog(DF_MOD, DS_ERROR, "Error when parsing directive time_field\n");
    return;
  }
  *pos = '\0';

  add_field(dir->args, pos + 1, mod->config, T_CTIME);
}

static void
dir_add_ipv4_field(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  char* pos;

  for (pos = dir->args; *pos && !isblank(*pos); pos++)
    continue;
  if (!pos)
  {
    DebugLog(DF_MOD, DS_ERROR, "Error when parsing directive ipv4_field\n");
    return;
  }
  *pos = '\0';

  add_field(dir->args, pos + 1, mod->config, T_IPV4);
}

static void
dir_add_int_field(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  char* pos;

  for (pos = dir->args; *pos && !isblank(*pos); pos++)
    continue;
  if (!pos)
  {
    DebugLog(DF_MOD, DS_ERROR, "Error when parsing directive int_field\n");
    return;
  }
  *pos = '\0';

  add_field(dir->args, pos + 1, mod->config, T_INT);
}

static void
add_udp_source(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  int *p;

  DebugLog(DF_MOD, DS_INFO, "Adding udp source port %s\n", dir->args);

  /* XXXXXX Rhaaa p is in the stack ! correct this rapidly !!!
  ** now it is an unreferenced int... ;-/ */
  p = Xmalloc(sizeof (int));
  *p = atoi(dir->args);
  register_conditional_dissector(ctx, mod, "udp",
                                 (void *)p, sizeof(int),
                                 dissect_idmef, NULL);
}

static void
add_textfile_source(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  DebugLog(DF_MOD, DS_INFO, "Adding textfile source file %s\n", dir->args);

  register_conditional_dissector(ctx, mod, "textfile",
                                 (void *)dir->args, strlen(dir->args),
                                 dissect_idmef, NULL);
}

static void
add_sockunix_source(orchids_t *ctx, mod_entry_t *mod, config_directive_t *dir)
{
  DebugLog(DF_MOD, DS_INFO, "Adding sockunix source %s\n", dir->args);

  register_conditional_dissector(ctx, mod, "sockunix",
                                 (void *)dir->args, strlen(dir->args),
                                 dissect_idmef, NULL);
}

static mod_cfg_cmd_t idmef_cfgcmds[] = {
  { "AddTextfileSource", add_textfile_source, "Add text source file" },
  { "AddUdpSource", add_udp_source, "Add udp port source" },
  { "AddUnixSocketSource", add_sockunix_source, "Add a unix socket source"},
  { "str_field", dir_add_str_field, "Add a new field corresponding to a XPath query."},
  { "time_field", dir_add_time_field, "Add a new field corresponding to a XPath query."},
  { "ipv4_field", dir_add_ipv4_field, "Add a new field corresponding to a XPath query."},
  { "int_field", dir_add_int_field, "Add a new field corresponding to a XPath query."},
  { NULL, NULL, NULL }
};


static char *idmef_deps[] = {
  "udp",
  "textfile",
  NULL
};


input_module_t mod_idmef = {
  MOD_MAGIC,
  ORCHIDS_VERSION,
  "idmef",
  "CeCILL2",
  idmef_deps,
  idmef_cfgcmds,
  idmef_preconfig,
  idmef_postconfig,
  NULL
};

/*
** Copyright (c) 2002-2005 by Julien OLIVAIN, Laboratoire Spécification
** et Vérification (LSV), CNRS UMR 8643 & ENS Cachan.
**
** Julien OLIVAIN <julien.olivain@lsv.ens-cachan.fr>
**
** This software is a computer program whose purpose is to detect intrusions
** in a computer network.
**
** This software is governed by the CeCILL license under French law and
** abiding by the rules of distribution of free software.  You can use,
** modify and/or redistribute the software under the terms of the CeCILL
** license as circulated by CEA, CNRS and INRIA at the following URL
** "http://www.cecill.info".
**
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
**
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
**
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL license and that you accept its terms.
*/

/* End-of-file */
